import logging
from collections import UserDict
from typing import List

from mev_yolodeep.config import config

from .detection import Detection, DetectionPerson
from .wamp import wamp

log = logging.getLogger(__name__)


class Collector(UserDict):
    def __init__(self, source_id=0):
        super().__init__()
        self.source_id = source_id

    def add(self, detections: List[Detection]):
        """
        all detections here have same timestamp and same frame

        """
        pgies = [
            detection
            for detection in detections
            if detection.parent_id is None and detection.valid
        ]

        frame = None
        if config.osd or config.show or config.save or config.rtsp:
            if pgies:
                first = pgies[0]
                frame = first.getFrame()

        for detection in pgies:
            if detection.object_id not in self:
                self[detection.object_id] = []

            self._check_position_and_add(detection)

            if config.osd:
                detection.osd_render()

            if frame is not None:  # questa parte è piuttosto lenta
                detection.draw(frame, older=self[detection.object_id][:-1])
            elif config.hide_persons and isinstance(detection, DetectionPerson):
                frame = (
                    detection.getFrame()
                )  # possiamo così riciclarlo per altre persone da oscurare
                detection.draw(frame)

        sgies = [
            detection for detection in detections if detection.parent_id is not None
        ]
        for sgie in sgies:
            parent = self._current(sgie.parent_id)
            if parent:
                parent.addChild(sgie)

        self._pollLive([detection.object_id for detection in pgies])

    def _check_position_and_add(self, detection):
        """
        aggiunge alle detections o rimpiazza l'ultima se la nuova è nella stessa posizione
        """
        if self[detection.object_id]:
            previous = self[detection.object_id][-1]
            distance = detection.get_distance(previous)
            if (
                distance <= config.min_distance
            ):  # se il nuovo è vicino sostituisci il precedente
                self[detection.object_id].pop()
        self[detection.object_id].append(detection)

    def _pollLive(self, ids: List[int]):
        """
        check live objects
        """
        for object_id in list(self):
            detection = self._current(object_id)
            if not detection:
                continue
            if object_id in ids:
                detection.seen()
            else:
                detection.notSeen()

            if detection.isLive():
                detection.analyze(brothers=self[object_id], live=True)
            else:
                detection.analyze(brothers=self[object_id])
                self.pop(object_id)

    def _current(self, object_id):
        try:
            return self[object_id][-1]
        except Exception:
            return None


class MainCollector:
    """
    one collector for all sources
    """

    def __init__(self):
        self.collectors = [
            Collector(source_id=0),  # source_id=0
            Collector(source_id=1),
        ]

    def add(self, detections: List[Detection], source_id: int = 0, frame_num: int = -1):
        if frame_num % 6 == 0:  # 5 frames al secondo
            wamp.publish(
                "frame.new",
                source_id=source_id,
                frame_num=frame_num,
                path=config.save_frames,
            )
        if detections:
            source_id = detections[0].source_id
            self.collectors[source_id].add(detections)

    def __repr__(self):
        return "\n".join([str(c) for c in self.collectors])

    def framesave(self, source_id=0, frame_num=1):
        self.collectors[source_id].framesave(frame_num)
