import json
import logging
import math
import os
import shutil
import statistics
from enum import IntEnum
from threading import Thread
from typing import List

import cv2 as cv
import Levenshtein
import numpy as np
import pyds

from mev_yolodeep.config import config

from .wamp import wamp

log = logging.getLogger(__name__)


"""
>
messaggio da inviare al brain

    async def saveAndPublish(self, publish):
        plate_ascii, plate = self.getBestValue()
        score = self.getBestScore() * 100
        key = "%s_%s" % (plate_ascii, self.randomString())

        medium_angle = 0
        parking = ParkingAngle.parallel
        if self.angles:
            medium_angle = sum(self.angles) / len(self.angles)
            parking = self.getParkingAngle(medium_angle, self.camera_id)

        payload = dict(
            key=key,
            plate_ascii=plate_ascii,
            score=score,
            plate=plate,
            timestamp=self.time,
            quarter=int(self.camera_id),
            x=self.centroid[0],
            y=self.centroid[1],
            w=int(self.box_best[2]),
            h=int(self.box_best[3]),
            angle=medium_angle,
            parking=parking,
        )
        self.toJpg(config, key)
        log.info(
            "ocr plate=%s score=%.1f angle=%.1f parking=%s",
            plate_ascii,
            score,
            payload["angle"],
            payload["parking"],
        )
        publish("plate.new", plate=payload)

"""


class ParkingAngle(IntEnum):
    parallel = 0
    vertical = 1


def getDetection(obj_meta, frame_meta, info=None):
    """
    0: plate
    1: car
    2: bus
    3: truck
    4: motorcycle
    5: bike
    6: driver
    7: person

    0: plate
    1: car
    2: motorcycle
    3: face
    """
    if obj_meta.unique_component_id == 1:
        if obj_meta.class_id == 0:
            return DetectionPlate(obj_meta, frame_meta, info=info)
        elif obj_meta.class_id == 3:
            return DetectionPerson(obj_meta, frame_meta, info=info)
        elif obj_meta.class_id in (1, 2):
            return DetectionVehicle(obj_meta, frame_meta, info=info)
        else:
            return DetectionPGIE(obj_meta, frame_meta, info=info)
    else:
        return DetectionOCR(obj_meta, frame_meta, info=info)


def initDetection(config):
    Detection.config = config


class Detection:
    """
    obj_meta arriva dal primary engine pgie(id=1) o dal secondario sgie(id=2)

    obj_meta.classifier_meta_list arriva da engine sgie(id=3)

    con lo stesso timestamp arrivano:
        object_id=18446744073709551615 parent=2 label=V-27(0.93)
        object_id=18446744073709551615 parent=2 label=M-21(0.93)
        object_id=18446744073709551615 parent=2 label=G-16(0.94)
        object_id=18446744073709551615 parent=2 label=C-12(0.93)
        object_id=18446744073709551615 parent=2 label=0-0(0.92)
        object_id=18446744073709551615 parent=2 label=0-0(0.92)
        object_id=18446744073709551615 parent=2 label=0-0(0.94)
        object_id=18446744073709551615 parent=4 label=M-21(0.62)
        object_id=18446744073709551615 parent=4 label=M-21(0.64)
        object_id=18446744073709551615 parent=4 label=4-4(0.54)
        object_id=1 parent=0 label=car-1(0.94)
        object_id=0 parent=0 label=car-1(0.94)
        object_id=3 parent=0 label=car-1(0.96)
        object_id=4 parent=0 label=plate-0(0.86)
        object_id=2 parent=0 label=plate-0(0.94)

    note: object_id sui secondary detector arriva costante, quindi lo metto a 0

    """

    config = None

    def __init__(self, obj_meta, frame_meta, info=None):
        self.info = info
        self.detector_id = obj_meta.unique_component_id  # detector id
        self.object_id = obj_meta.object_id if self.detector_id == 1 else None
        self.class_id = obj_meta.class_id
        self.confidence = max(
            0.0, round(obj_meta.confidence, 2)
        )  # a volte arriva una confidenza negativa !
        self.frame_num = frame_meta.frame_num
        self.batch_id = frame_meta.batch_id
        self.timestamp = frame_meta.ntp_timestamp / 1e9
        self.frame_height = frame_meta.source_frame_height
        self.frame_width = frame_meta.source_frame_width
        self.label = obj_meta.obj_label
        self.source_id = frame_meta.source_id
        self.box = self._get_box(obj_meta.rect_params)
        self.centroid = self.get_centroid()
        self.classification_class_id = 0
        self.classification_confidence = 0.0

        self.parent_id = None
        self.parent_bbox = None
        if obj_meta.parent:
            self.parent_id = obj_meta.parent.object_id
            self.parent_bbox = obj_meta.parent.detector_bbox_info.org_bbox_coords
        if obj_meta.classifier_meta_list is not None:
            self._parse_classification(obj_meta)
        self.valid = self.is_valid()
        # self._render_osd(obj_meta)

        self.children: List[Detection] = []
        self.frame_num_last = self.frame_num
        self.not_seen = 0
        if (
            self.config.osd
        ):  # ci potrebbero essere dei memory leaks, quindi la limitiamo al debug
            self.obj_meta = obj_meta
        else:
            self.obj_meta = None

    def osd_render(self):
        """
        dopo il render togliamo i riferimenti a obj_meta per evitare leaks
        """

        if self.obj_meta:
            self._render_osd(self.obj_meta)
            self.obj_meta = None

    def getFrame(self):
        gst_buffer = self.info.get_buffer()
        return pyds.get_nvds_buf_surface(hash(gst_buffer), self.batch_id)

    def addChild(self, child):
        self.children.append(child)
        self.timestamp = child.timestamp  # keep alive

    def isLive(self, frame_not_seen=2):
        return self.not_seen <= frame_not_seen

    def notSeen(self):
        self.not_seen += 1

    def seen(self):
        self.not_seen = 0

    def _parse_classification(self, obj_meta):
        l_classifier = obj_meta.classifier_meta_list
        while l_classifier is not None:
            class_meta = pyds.NvDsClassifierMeta.cast(l_classifier.data)
            if class_meta is not None:
                label_meta = class_meta.label_info_list
                while label_meta is not None:
                    label_info = pyds.NvDsLabelInfo.cast(label_meta.data)
                    # showObject(label_info)
                    self.classification_class_id = label_info.result_class_id
                    self.classification_confidence = round(label_info.result_prob, 2)
                    try:
                        label_meta = label_meta.next
                    except StopIteration:
                        break
            try:
                l_classifier = l_classifier.next
            except StopIteration:
                break

    def _render_osd(self, obj_meta):
        pass

    def __repr__(self):
        return f"{self.__class__.__name__} source={self.source_id} frame={self.frame_num} frame_last={self.frame_num_last} object_id={self.object_id} not_seen={self.not_seen} obj={self.label}-{self.class_id}({self.confidence:.2f})"

    def draw(self, frame, older=None):
        return frame

    def _get_box(self, rect_params):
        return (
            int(rect_params.left),
            int(rect_params.top),
            int(rect_params.width),
            int(rect_params.height),
        )

    def _draw_trajectory(self, image, older=None):
        blk = np.zeros(image.shape, np.uint8)
        old = self._draw_centroid(blk, self)
        if older:
            for i, detection in enumerate(reversed(older)):
                b = max(0, 255 - 10 * (i + 1))
                color = (0, b, b)
                current = self._draw_centroid(blk, detection, color=color, past=i + 1)
                self._draw_line(blk, old, current, color=color)
                old = current
        image[:] = cv.addWeighted(image, 1.0, blk, 0.8, 1)

    def get_centroid(self, detection=None):
        if detection is None:
            detection = self
        left, top, width, height = detection.box
        x = int(left + width / 2)
        y = int(top + height / 2)
        return (x, y)

    def get_distance(self, detection):
        """get distance perc from other detection"""

        x1, y1 = self.centroid  # get_centroid()
        x2, y2 = detection.centroid  # get_centroid()

        x1 = x1 / self.frame_width
        x2 = x2 / self.frame_width
        y1 = y1 / self.frame_height
        y2 = y2 / self.frame_height

        return math.sqrt((y2 - y1) ** 2 + (x2 - x1) ** 2)

    def _draw_centroid(self, image, detection, color=(0, 255, 255), past=0):
        # x, y = self.get_centroid(detection)
        x, y = detection.centroid
        cv.circle(image, (x, y), 10 if past else 35, color, cv.FILLED)
        if past == 0:
            cv.putText(
                image,
                str(self.object_id),
                (x - 27, y + 8),
                cv.FONT_HERSHEY_SIMPLEX,
                1.0,
                (255, 255, 255),
                2,
                cv.LINE_AA,
            )

        return (x, y)

    # def drawInfo(self, image):
    #     label = str(self.frame_num)
    #     x1,y1,w,h = self.frame_width,0,60,30
    #     x1 = x1 - w
    #     cv.rectangle(image, (x1, y1), (x1 + w, y1+h), (0,0,0), -1)
    #     cv.putText(image, label, (x1+3, y1+h-4), cv.FONT_HERSHEY_SIMPLEX, 1.0, (255,255,255), 2, cv.LINE_AA)

    def _draw_line(self, blk, point1, point2, color=(0, 255, 255)):
        cv.line(blk, point1, point2, color, 2)

    def is_valid(self, detections=[]):
        return True

    def _anonymize_bbox(self, image, mode="pixelate"):
        left, top, width, height = self.box

        x1 = left
        y1 = top
        x2 = left + width
        y2 = top + height

        if mode == "blur":
            bbox = image[y1:y2, x1:x2]
            image[y1:y2, x1:x2] = cv.GaussianBlur(bbox, (15, 15), 60)
        elif mode == "pixelate":
            reshape_factor = 18
            min_dim = 16
            bbox = image[y1:y2, x1:x2]
            h, w, _ = bbox.shape
            new_shape = (
                max(min_dim, int(w / reshape_factor)),
                max(min_dim, int(h / reshape_factor)),
            )
            bbox = cv.resize(bbox, new_shape, interpolation=cv.INTER_LINEAR)
            image[y1:y2, x1:x2] = cv.resize(
                bbox, (w, h), interpolation=cv.INTER_NEAREST
            )
        elif mode == "fill":
            image = cv.rectangle(
                image, (x1, y1), (x2, y2), color=(0, 0, 0), thickness=-1
            )
        else:
            raise ValueError(f"Invalid anonymization mode '{mode}'.")

        return image

    def analyze(self, brothers, n_samples=5, live=False):
        """overload me"""
        return None

    def save(self, outdir="/tmp"):
        frame = self.getFrame()
        frame_copy = np.array(frame, copy=True, order="C")
        frame_copy = cv.cvtColor(frame_copy, cv.COLOR_RGBA2BGR)
        if not os.path.exists(outdir):
            os.makedirs(outdir)

        filename = os.path.join(
            outdir, f"{self.frame_num}_{self.source_id}_{self.object_id}.jpg"
        )
        cv.imwrite(filename, frame_copy)
        log.info("frame saved to %s", filename)

    def save_context(self, plate):
        if config.save_async:
            t1 = Thread(target=self._save_context_blocking, args=(plate, True))
            t1.start()
        else:
            self._save_context_blocking(plate)

    def _save_context_blocking(self, plate, is_thread=False):
        """
        save context frame from config.save_frames to config.save_context
        """
        left, top, width, height = plate.box

        frame_filename = self._get_frame_name(plate)
        context_filename = f"{config.save_context}/{plate.source_id}_{plate.frame_num:07d}_{left}-{top}-{width}-{height}_{plate.nationality}_{plate.license}"

        try:
            shutil.copyfile(frame_filename, context_filename + ".jpg")
            payload = {
                "frame_num": plate.frame_num,
                "source_id": plate.source_id,
                "plate_box": plate.box,
                "nationality": plate.nationality,
                "license": plate.license,
            }
            with open(context_filename + ".json", "w") as f:
                json.dump(payload, f)

            log.info(
                " save_context: source_id=%s frame_num=%d thread=%s saving from %s to %s.*",
                plate.source_id,
                plate.frame_num,
                is_thread,
                frame_filename,
                context_filename,
            )
        except Exception:
            log.warning(
                " save_context cannot save: source_id=%s frame_num=%d saving from %s to %s.*",
                plate.source_id,
                plate.frame_num,
                frame_filename,
                context_filename,
            )
            return

    def on_destroy(self):
        pass


class DetectionPGIE(Detection):
    def _render_osd(self, obj_meta):
        obj_meta.rect_params.has_bg_color = True
        obj_meta.rect_params.border_color.set(0.0, 0.0, 1.0, 0.5)
        obj_meta.rect_params.bg_color.set(0.0, 0.0, 1.0, 0.15)
        obj_meta.rect_params.border_width = 2

        obj_meta.text_params.font_params.font_size = int(
            config.zoom * 24 / (config.num_sources or 1)
        )
        obj_meta.text_params.display_text = (
            f"{self.label} {str(obj_meta.object_id)[-4:]}"
        )


class DetectionPlate(DetectionPGIE):
    _min_chars = 7
    _max_chars = 8

    def __init__(self, obj_meta, frame_meta, info=None):
        super().__init__(obj_meta, frame_meta, info=info)
        self.license = ""
        self.score = 0.0
        self.nationality = self.get_nationality()
        self.angle = 0
        self.parking = ParkingAngle.parallel
        self.notified = False

    def __repr__(self):
        msg = super().__repr__()
        msg += f" classification={self.classification_class_id}({self.classification_confidence:.2f})"
        return msg

    def _render_osd(self, obj_meta):
        super()._render_osd(obj_meta)
        obj_meta.rect_params.has_bg_color = True
        if self.valid:
            obj_meta.rect_params.bg_color.set(0.0, 1.0, 0.0, 0.2)
            obj_meta.rect_params.border_color.set(0.0, 1.0, 0.0, 0.5)
        else:
            obj_meta.rect_params.bg_color.set(1.0, 0.0, 0.0, 0.2)
            obj_meta.rect_params.border_color.set(1.0, 0.0, 0.0, 0.5)
        obj_meta.rect_params.border_width = 3

        nationality = self.get_nationality()
        obj_meta.text_params.display_text = (
            f"{nationality} {self.classification_confidence} {str(self.object_id)[-4:]}"
        )

    def get_nationality(self):
        if self.classification_class_id == 5 and self.classification_confidence >= 0.5:
            return "ITA"
        return "EE"

    def draw(self, frame, older=None):
        if config.privacy:
            frame = self._anonymize_bbox(frame)
        return frame

    def seen(self):
        super().seen()
        self._sort_children()
        self.license = "".join([child.label for child in self.children])
        self.score = self._calculate_score()

    def is_valid(self, detections=[]):
        """
        le targhe della camera anteriore vengono viste con aree sempre maggiori

        """
        # if not detections:
        #     return True

        margin = 5  # 5 pixel di margine dai bordi
        left, top, width, height = self.box
        # log.info(
        #     "is_valid: source_id=%d object_id=%s left=%d frame_width=%d",
        #     self.source_id,
        #     self.object_id,
        #     left,
        #     self.frame_width,
        # )
        if self.source_id == 0:
            ## le targhe sulla camera anteriore non devono essere troncate dal bordo destro
            # log.info("is_valid: object_id=%s left+width=%d frame_width=%d", self.object_id, left+width, self.frame_width)
            if (
                left + width + margin > self.frame_width
            ):  # se la targa è troncata sul bordo destro
                return False
            # current_width = self.box[2]
            # previuos_width = detections[-1].box[2]
            # log.info("is_valid: object_id=%s  previous=%d new=%d", self.object_id, previuos_width, current_width)
            # if current_width < previuos_width:
            #     return False
        elif self.source_id == 1:
            if left <= margin:
                return False
        return True

    def _calculate_score(self):
        score = 0.0
        if self.children:
            if self._min_chars <= len(self.children) <= self._max_chars:
                # score = min([child.confidence for child in self.children])
                # score = statistics.median([child.confidence for child in self.children])
                score = statistics.mean([child.confidence for child in self.children])
        return score

    def _sort_children(self):
        if not self.children:
            return

        detections = []
        boxes = self._sort_boxes([child.box for child in self.children])
        for box in boxes:
            for detection in self.children:
                if np.array_equal(detection.box, box):
                    detections.append(detection)
        self.children = detections

    def _sort_boxes(self, boxes):
        """
        left, top, width, height = box

        ordina i punti secondo x e poi
        assegna a una riga diversa ogni volta che delta y supera +soglia oppure -soglia

        soglia è i 0.6 dell'altezza media dei simboli
        """
        _SOGLIA_Y = 0.6
        if not boxes:
            return []

        def __sort_x(box):
            return box[0]

        boxes = sorted(boxes, key=__sort_x)

        hlimit = np.mean(np.array([box[3] for box in boxes])) * _SOGLIA_Y

        rows = {0: [boxes[0]]}
        current_row = 0
        for box in boxes[1:]:
            previous_y = rows[current_row][-1][1]
            deltay = box[1] - previous_y
            if deltay > hlimit:  # quando il delta y supera la soglia cambio riga
                current_row += 1
            elif deltay < -hlimit:  # quando il delta è negativo cambio riga
                current_row -= 1
            if current_row not in rows:  # aggiungo una riga se necessario
                rows[current_row] = []
            rows[current_row].append(box)
        ordered = [rows[key] for key in sorted(rows.keys())]
        return np.concatenate(ordered)

    def on_destroy(self):
        pass  # log.info("license detected %s", self.license)

    def analyze(self, brothers, n_samples=5, live=False):
        """
        abbiamo una serie di targhe dello stesso object_id ordinate secondo il tempo
        brothers = [first_plate, plate2, plate3, ...]

        se la camera è anteriore notifichiamo quando live=False, cioè quando l'oggetto
        non è più visto dal tracker

        first_plate è la targa di riferimento di tutti i fratelli, dove ad esempio salviamo se è stata notificata

        """

        plates = [
            plate for plate in brothers if plate.valid and plate.score > 0
        ]  # keep only valid plates
        if not plates:
            return

        first_plate = brothers[0]
        if first_plate.notified:
            return

        # if self.source_id == 0:
        #     # la telecamera frontale riceve targhe valide solo alla fine
        #     plates = plates[-n_samples:]
        # else:
        #     # mentre la camera posteriore riceve targhe valide all'inizio
        #     plates = plates[:n_samples]
        log.debug(
            "plates(object_id=%d) source=%d-%d:",
            self.object_id,
            self.source_id,
            self.frame_num,
        )
        plates = sorted(plates, key=lambda parent: parent.score, reverse=True)
        best_plates = self._filter_plates_by_license(plates)[:n_samples]
        for instance in best_plates:
            log.debug(
                "  plate license=%s score=%.2f nationality=%s",
                instance.license,
                instance.score,
                instance.nationality,
            )

        notify = False
        if not live:
            notify = True

        if self.source_id == 1 and live:
            last_plate = brothers[-1]
            first_frame_num = first_plate.frame_num
            i = 0
            while i < len(brothers):
                if brothers[i].score > 0:
                    first_frame_num = brothers[i].frame_num
                    break
                i += 1
            delta_frame_num = last_plate.frame_num - first_frame_num
            log.debug(
                "delta_frame_num=%d len(best_plates)=%d",
                delta_frame_num,
                len(best_plates),
            )

            # se ho un numero sufficiente di osservazioni
            if (
                delta_frame_num > len(best_plates) + 3
            ):  # se minore di 3 a volte le immagini da salvare non sono ancora presenti su filesystem
                notify = True

        if notify:
            best = best_plates[
                0
            ]  # prendiamo come riferimento la targa con score maggiore
            best.nationality = self._get_best_nationality(best_plates)
            best.license = self._get_best_license(best_plates)
            best.angle = self._get_best_angle(best_plates)
            best.parking = best.get_parking_angle()
            best.frame_num = self._get_frame_num(plates, brothers[-1].frame_num)
            log.info(
                " --> plate(%d) source=%d-%d license=%s score=%.2f score-min=%.2f nation=%s angle=%.1f",
                best.object_id,
                best.source_id,
                best.frame_num,
                best.license,
                best.score,
                config.score_min,
                best.nationality,
                best.angle,
            )
            if best.score >= config.score_min:
                if config.save_frames and config.save_context:
                    self.save_context(best)
                self._publish_wamp(best)
            first_plate.notified = True
            return best

    def _get_frame_name(self, plate):
        return f"{config.save_frames}/{plate.source_id}_{plate.frame_num:07d}.jpg"

    def _get_frame_num(self, plates, frame_num_current):
        """
        Cerchiamo il frame di contesto ancora presente su disco, le targhe sono ordinate per score.
        Troviamo la targa tra le migliori che hanno un frame_num non troppo distante dal corrente
        """
        frame_num_min = max(1, frame_num_current - 2 * self.config.max_frames / 3)
        log.debug(
            "_get_frame_num: frame_num_current=%d frame_num_min=%d",
            frame_num_current,
            frame_num_min,
        )
        for plate in plates:
            if plate.frame_num >= frame_num_min:
                # potrei arrivare qui con frames non ancora scritti su file oppure già eliminati
                if os.path.exists(self._get_frame_name(plate)):
                    return plate.frame_num
        log.warning(
            "no context image found after %d for object_id=%d",
            frame_num_min,
            plates[0].object_id,
        )
        return 0

    def _filter_plates_by_license(self, plates):
        """
        filtriamo solo le targhe che hanno il numero di caratteri più frequente
        """
        n_chars_most_frequent = statistics.mode(
            [len(plate.license) for plate in plates]
        )
        return [
            plate for plate in plates if len(plate.license) == n_chars_most_frequent
        ]

    def _get_best_angle(self, plates):
        return statistics.mean([plate._get_angle() for plate in plates])
        # medium_angle = 0
        # if self.angles:
        #     medium_angle = sum(self.angles) / len(self.angles)
        #     parking = self.getParkingAngle(medium_angle, self.camera_id)

    def _publish_wamp(self, plate):
        # x, y = plate.get_centroid()
        x, y = plate.centroid
        _, _, width, height = self.box
        payload = {
            "key": f"{plate.source_id}_{plate.frame_num:07d}",
            "object_id": plate.object_id,
            "plate_ascii": plate.license,
            "score": plate.score,
            "plate": plate.license,
            "timestamp": plate.timestamp,
            "quarter": int(plate.source_id),
            "x": x,
            "y": y,
            "w": width,
            "h": height,
            "angle": plate.angle,  # medium_angle,
            "parking": int(plate.parking),
            "nationality": plate.nationality,
        }
        wamp.publish("plate.new", plate=payload)

    def _get_best_nationality(self, plates):
        try:
            return statistics.mode([plate.nationality for plate in plates])
        except Exception:
            return plates[0].nationality

    def _get_best_license(self, plates):
        return Levenshtein.median([plate.license for plate in plates])

    def _get_angle(self):
        """interpolate center (x,y) of each character boxes with y = mx + c"""
        if len(self.children) <= 1:  # non possiamo desumere nulla
            return 0
        x = np.array([child.centroid[0] for child in self.children])
        y = np.array([child.centroid[1] for child in self.children])
        A = np.vstack([x, np.ones(len(x))]).T
        m, c = np.linalg.lstsq(A, y, rcond=None)[0]
        return -int(round(math.degrees(m)))  # inverti l'angolo perché y parte dall'alto

    def get_parking_angle(self) -> ParkingAngle:
        if self.source_id in (0, 2) and self.angle >= -config.plate_angle_vertical:
            return ParkingAngle.parallel
        if self.source_id in (1, 3) and self.angle <= config.plate_angle_vertical:
            return ParkingAngle.parallel
        return ParkingAngle.vertical


class DetectionOCR(Detection):
    def __repr__(self):
        return f"SGIE source={self.source_id} frame={self.frame_num} parent={self.parent_id} obj={self.label}-{self.class_id}({self.confidence:.2f})"

    def _render_osd(self, obj_meta):
        obj_meta.text_params.text_bg_clr.set(1.0, 0.0, 0.0, 0.6)
        obj_meta.text_params.font_params.font_size = int(
            config.zoom * 24 / (config.num_sources or 1)
        )
        y = self.parent_bbox.top + self.parent_bbox.height
        obj_meta.text_params.y_offset = int(y)  # spostiamo i caratteri in basso

        obj_meta.rect_params.has_bg_color = False
        # obj_meta.rect_params.border_color.set(1.0, 0.0, 0.0, 0.3)
        obj_meta.rect_params.bg_color.set(1.0, 0.0, 0.0, 0.8)
        obj_meta.rect_params.border_width = 2

        if config.privacy:
            obj_meta.text_params.display_text = "X"


class DetectionPerson(Detection):
    def draw(self, frame, older=None):
        """in ogni caso pixeliamo la persona, in futuro di potrebbe fare a livello osd"""

        # if config.privacy: # or not config.save:
        frame = self._anonymize_bbox(frame)
        return frame

    def _render_osd(self, obj_meta):
        super()._render_osd(obj_meta)
        obj_meta.text_params.display_text = "person"
        obj_meta.rect_params.has_bg_color = False
        # obj_meta.rect_params.has_bg_color = True
        # obj_meta.rect_params.bg_color.set(0.0, 0.0, 0.0, 0.9)
        obj_meta.rect_params.border_color.set(0.0, 0.0, 0.0, 0.3)
        obj_meta.rect_params.border_width = 2


class DetectionVehicle(DetectionPGIE):
    def draw(self, frame, older=None):
        frame = self._draw_trajectory(frame, older=older)
        return frame

    def _render_osd(self, obj_meta):
        super()._render_osd(obj_meta)
        if not self.valid:
            obj_meta.rect_params.has_bg_color = False
            obj_meta.rect_params.border_color.set(1.0, 0.0, 0.0, 0.3)
            obj_meta.text_params.display_text = ""

    def is_valid(self, detections=[]) -> bool:
        margin_x = 0.25
        margin_y = 0.5
        left, top, width, height = self.box
        x, y = self.centroid  # get_centroid()

        if self.source_id == 0:
            is_bottom_left = (
                y > margin_y * self.frame_height and x < margin_x * self.frame_width
            )
            return not is_bottom_left
        else:
            is_bottom_right = (
                y > margin_y * self.frame_height
                and x > (1 - margin_x) * self.frame_width
            )
            return not is_bottom_right
