"""This module contains all Gst pipeline related logic."""

import asyncio
import logging
import math
import os
# import socket
# import struct
import time
from collections import defaultdict
from functools import partial
from inspect import signature
from typing import List

import asset
import cv2
import gi
import numpy as np
import pyds

from mev_yolodeep.lib.detection import Detection, getDetection, initDetection

from .collector import MainCollector
from .wamp import init_wamp, wamp
from .utils.bbox import rect_params_to_coords
from .utils.fps import FPSMonitor
from .utils.is_aarch_64 import is_aarch64

gi.require_version("Gst", "1.0")
gi.require_version("GstRtspServer", "1.0")
from gi.repository import GLib, GObject, Gst  # noqa: E402

log = logging.getLogger(__name__)

OUTPUT_DIR = "/tmp"
CROPS_DIR = "/tmp"


def quit(loop):
    if getattr(loop, "stop", None):
        loop.stop()
    else:
        loop.quit()  # da rimuovere quando wamp sarà attivo


class Pipeline:
    counter = 1

    def __init__(
        self,
        config=None,
        pgie_config_path: str = os.path.join("pgies/pgie.txt"),
        tracker_config_path: str = os.path.join("trackers/nvdcf.txt"),
        enable_osd: bool = True,
        write_osd_analytics: bool = True,
        save_crops: bool = False,
        # output_format: str = "mp4",
    ):
        """Create a Deepstream Pipeline.

        Args:
            video_uri (str): The source video URI.
            pgie_config_path (str): The configuration file path of the primary inference engine.
            tracker_config_path (str): The configuration file path of the tracker.
            enable_osd (bool): Whether to enable the on-screen display.
            write_osd_analytics (bool): Write analytics to the OSD.
            save_crops (bool): Save the object image crops to disk.
            output_format (str): The Deepstream output format.
            codec (str): The RTSP output video codec.
            input_shape (tuple): The video input shape (width, height).
        """
        self.config = config
        initDetection(config)
        self.logger = log  # logging.getLogger(__name__ + "." + self.__class__.__name__)
        self.pgie_config_path = pgie_config_path
        self.tracker_config_path = tracker_config_path
        self.enable_osd = enable_osd
        self.write_osd_analytics = write_osd_analytics
        self.save_crops = save_crops
        # self.output_format = output_format
        self.num_sources = 0
        self.fps_streams = {}
        self.udp_sink_port = 5400
        self.collector = None

        # for i in range(self.num_sources):
        #     self.fps_streams[f"stream{i}"] = FPSMonitor(i)

        if self.save_crops:
            if not os.path.isdir(CROPS_DIR):
                os.makedirs(CROPS_DIR)
            self.logger.info(f"Saving crops to '{CROPS_DIR}'.")
            self.track_scores = defaultdict(list)

        # self.logger.info(f"Playing from URI {self.video_uri}")
        GObject.threads_init()
        Gst.init(None)

        self.logger.info("Creating Pipeline")
        self.pipeline = Gst.Pipeline()
        if not self.pipeline:
            self.logger.error("Failed to create Pipeline")

        self.elements = []
        self.source_bin = None
        self.streammux = None
        self.pgie = None
        self.tracker = None
        # self.nvvidconv1 = None
        self.capsfilter1 = None
        self.tiler = None
        # self.nvvidconv2 = None
        self.nvosd = None
        self.queue1 = None
        # self.sink_bin = None

        self.elements_out = []
        self.cameras = {}

        self.setup()
        self._link_elements()
        # self.wampclient = None
        # self._add_probes()
        # self.wamp = None
        # if self.config.wamp:
        #     self.logger.info("creating wamp runner")
        #     self.wamp = ApplicationRunner(
        #         url=config.ws, realm=config.realm, serializers=[JsonSerializer()]
        #     )
        # if self.config.gps_sync:  # per il debug
        #     log.info("enabling gps sync signal on port %s", self.config.socket_port_gps)
        #     self.sock_emit = socket.socket(
        #         socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP
        #     )

    def __str__(self):
        return " -> ".join([elm.name for elm in self.elements])

    def bus_call(self, bus, message, loop):
        t = message.type
        if t == Gst.MessageType.EOS:
            log.debug("bus_call: end-of-stream")
            quit(loop)
        elif t == Gst.MessageType.WARNING:
            err, debug = message.parse_warning()
            log.warning("%s: %s", err, debug)
        elif t == Gst.MessageType.ERROR:
            err, debug = message.parse_error()
            log.error("%s: %s", err, debug)
            quit(loop)
        elif t == Gst.MessageType.ELEMENT:
            if "GstMultiFileSink" == message.src.__class__.__name__:
                structure = message.get_structure()
                self.onMultiFileSink(structure)

        return True

    def onMultiFileSink(self, structure):
        """
        callback called on jpg file saving (multifilesink)

        structure:
        filename=(string)/samples/xxx/00209.jpg, index=(int)209, timestamp=(guint64)6966666597, stream-time=(guint64)6966666597, running-time=(guint64)6966666597, duration=(guint64)18446744073709551615, offset=(guint64)18446744073709551615, offset-end=(guint64)18446744073709551615

        'can_intersect', 'copy', 'filter_and_map_in_place', 'fixate', 'fixate_field', 'fixate_field_boolean', 'fixate_field_nearest_double', 'fixate_field_nearest_fraction', 'fixate_field_nearest_int', 'fixate_field_string', 'foreach', 'free', 'from_string', 'get_array', 'get_boolean', 'get_clock_time', 'get_date', 'get_date_time', 'get_double', 'get_enum', 'get_field_type', 'get_flagset', 'get_fraction', 'get_int', 'get_int64', 'get_list', 'get_name', 'get_name_id', 'get_string', 'get_uint', 'get_uint64', 'get_value', 'has_field', 'has_field_typed', 'has_name', 'id_get_value', 'id_has_field', 'id_has_field_typed', 'id_set_value', 'id_take_value', 'intersect', 'is_equal', 'is_subset', 'keys', 'map_in_place', 'n_fields', 'name', 'new_empty', 'new_from_string', 'new_id_empty', 'nth_field_name', 'remove_all_fields', 'remove_field', 'set_array', 'set_list', 'set_name', 'set_parent_refcount', 'set_value', 'take_value', 'to_string', 'type'
        """
        return
        context_id = structure.get_value("index")
        log.debug("+++++ context_id=%d", context_id)
        # if self.collector:
        #     self.collector.addContext(context_id=context_id)

    def _add_element(self, element, idx=None):
        if idx:
            self.elements.insert(idx, element)
        else:
            self.elements.append(element)

        self.pipeline.add(element)

    def create_element(
        self, factory_name, name="", print_name="", add=True, properties=None
    ):
        """Creates an element with Gst Element Factory make.

        Return the element if successfully created, otherwise print to stderr and return None.
        """
        if not name:
            name = f"{factory_name}_{Pipeline.counter}"
            Pipeline.counter += 1
        if not print_name:
            print_name = name
        self.logger.debug("Creating %s", name)
        elm = Gst.ElementFactory.make(factory_name, name)

        if not elm:
            self.logger.error("Unable to create %s", name)
            return None

        if add:
            self._add_element(elm)

        if (
            factory_name in ("nvvideoconvert", "nvmultistreamtiler")
            and not is_aarch64()
        ):
            # Use CUDA unified memory so frames can be easily accessed on CPU in Python.
            elm.set_property("nvbuf-memory-type", int(pyds.NVBUF_MEM_CUDA_UNIFIED))

        if properties:
            for prop, value in properties.items():
                elm.set_property(prop, value)

        return elm

    def _element(self, factory_name, name="", properties=None, nolink=True):
        if not name:
            name = f"{factory_name}_{Pipeline.counter}"
            Pipeline.counter += 1
        self.logger.info("Creating %s", name)
        elm = Gst.ElementFactory.make(factory_name, name)
        if not elm:
            self.logger.error("Unable to create %s", name)
            return None
        if factory_name in ("nvvideoconvert", "nvmultistreamtiler"):
            # Use CUDA unified memory so frames can be easily accessed on CPU in Python.
            elm.set_property("nvbuf-memory-type", int(pyds.NVBUF_MEM_CUDA_UNIFIED))
        if factory_name in ("fakesink",):
            self.elements_out.append(elm)

        if properties:
            for prop, value in properties.items():
                elm.set_property(prop, value)
        if nolink:
            self.pipeline.add(elm)
        else:
            self._add_element(elm)

        return elm

    def _configure_basler_min_gain(
        self, pylon, index=0, padding_lateral=272, padding_top=258
    ):
        """
        stream::MaxTransferSize=4194304 \
        stream::MaxNumBuffer=10 \
        cam::AutoFunctionROIOffsetY-ROI1=258 \
        cam::AutoFunctionROIHeight-ROI1=1800 \
        cam::AutoFunctionProfile=MinimizeGain \
        cam::GainAuto=Continuous \
        cam::ExposureTime=4000 \
        cam::AutoFunctionROIUseBrightness-ROI1=1 \
        cam::AutoFunctionROIHighlight-ROI1=1


        l'immagine viene ritagliata in alto e lateralmente (a sx per la camera 0 e a dx per la 1)
        poi viene definito un ROI escludendo la fetta top
        """
        # l'ordine di questi parametri è importante
        stream = pylon.get_child_by_name("stream")
        stream.set_property("MaxTransferSize", 4194304)
        stream.set_property("MaxNumBuffer", 10)
        cam = pylon.get_child_by_name("cam")
        self.cameras[index] = cam
        cam.set_property("AutoFunctionProfile", "MinimizeGain")  # MinimizeExposureTime
        cam.set_property("GainAuto", "Continuous")
        cam.set_property("ExposureTime", self.config.exposure)  # microseconds
        cam.set_property("AutoFunctionROIUseBrightness-ROI1", 1)
        if self.config.roi:
            cam.set_property("AutoFunctionROIHighlight-ROI1", 1)
        ## questa non funziona e invece usiamo cameras/default.pfs solo per caricare questa impostazione
        # cam.set_property("AutoTargetBrightness", 0.5)

        cam.set_property("OffsetY", padding_top)

        if index == 1:
            cam.set_property("ReverseX", 1)
            cam.set_property("ReverseY", 1)
            cam.set_property("OffsetX", 0)
        else:
            cam.set_property("OffsetX", padding_lateral)

        cam.set_property("AutoFunctionROIHeight-ROI1", self.config.height - padding_top)
        cam.set_property(
            "AutoFunctionROIOffsetY-ROI1", self.config.height - padding_top
        )

        ## entuale roi con esclusione della parte laterale
        # cam.set_property("AutoFunctionROIWidth-ROI1", width-padding_lateral)
        # cam.set_property("AutoFunctionROIOffsetX-ROI1", 0)

    def _configure_basler_min_exposure(
        self, pylon, index=0, padding_lateral=272, padding_top=258
    ):
        """
        l'immagine viene ritagliata in alto e lateralmente (a sx per la camera 0 e a dx per la 1)
        poi viene definito un ROI escludendo la fetta top

        """
        log.info(
            "configure basler min_exposure index=%d target=%.2f gain_max=%.2f",
            index,
            self.config.target_brightness,
            self.config.gain_max,
        )
        # l'ordine di questi parametri è importante
        stream = pylon.get_child_by_name("stream")
        stream.set_property("MaxTransferSize", 4194304)
        stream.set_property("MaxNumBuffer", 10)
        cam = pylon.get_child_by_name("cam")
        self.cameras[index] = cam

        # cam.set_property("ExposureTime", 8439.0)  # non dovrebbe importare
        cam.set_property("ExposureAuto", "Continuous")
        cam.set_property("AutoFunctionProfile", "MinimizeExposureTime")
        # cam.set_property("BslAcquisitionStopMode", "CompleteExposure")
        cam.set_property("GainAuto", "Continuous")
        cam.set_property("AutoGainUpperLimit", self.config.gain_max)
        cam.set_property("AutoTargetBrightness", self.config.target_brightness)

        cam.set_property("AutoGainLowerLimit", 0.0)
        cam.set_property("AutoExposureTimeLowerLimit", 1.0)
        cam.set_property("AutoExposureTimeUpperLimit", 14000.0)

        # cam.set_property("BalanceWhiteAuto", "Continuous")
        cam.set_property("AutoFunctionROIUseBrightness-ROI1", 1)
        cam.set_property("AutoFunctionROIUseBrightness-ROI2", 0)
        # cam.set_property("AutoFunctionROIUseWhiteBalance-ROI1", 0)
        # cam.set_property("", "")

        # for prop in "ExposureTime,ExposureAuto,BslSensorBitDepthMode,AutoTargetBrightness,AutoFunctionProfile,GainAuto,BalanceWhiteAuto,AutoFunctionROIUseBrightness-ROI1".split(","):
        #     try:
        #         print(prop, cam.get_property(prop))
        #     except:
        #         print(f"no {prop} present")

        if self.config.roi:
            cam.set_property("AutoFunctionROIHighlight-ROI1", 1)
            cam.set_property("AutoFunctionROIHighlight-ROI2", 1)

        cam.set_property("OffsetY", padding_top)

        if index == 1:
            cam.set_property("ReverseX", 1)
            cam.set_property("ReverseY", 1)
            cam.set_property("OffsetX", 0)
        else:
            cam.set_property("OffsetX", padding_lateral)

        cam.set_property("AutoFunctionROIHeight-ROI1", self.config.height - padding_top)
        cam.set_property(
            "AutoFunctionROIOffsetY-ROI1", self.config.height - padding_top
        )

        ## entuale roi con esclusione della parte laterale
        # cam.set_property("AutoFunctionROIWidth-ROI1", width-padding_lateral)
        # cam.set_property("AutoFunctionROIOffsetX-ROI1", 0)

    def create_basler(self, index=0, serial=""):
        """
        gst-launch-1.0 pylonsrc capture-error=skip cam::ExposureTime=20000 cam::Gain=10.3 \
            ! "video/x-raw(memory:NVMM),width=640,height=480,framerate=10/1, format=GRAY8" \
            ! videoconvert \
            ! nvvideoconvert \
            ! m.sink_0 nvstreammux name=m batch-size=1 width=640 height=480 \
            ! nvinfer config-file-path=/home/arun/Documents/DeepStream-Yolo/config_infer_primary_yoloV5.txt ! nvdsosd ! nvegltransform ! nveglglessink sync=0

        """
        bin_name = "source-bin-%02d" % index
        nbin = Gst.Bin.new(bin_name)
        name = "pylonsrc-bin-%02d" % index
        pylon = Gst.ElementFactory.make("pylonsrc", name)

        pylon.set_property("device-serial-number", serial)
        # pylon.set_property(
        #     "pfs-location", asset.load("mev_yolodeep:assets/cameras/default.pfs").filename
        # )
        pylon.set_property("capture-error", "skip")
        Gst.Bin.add(nbin, pylon)
        self._configure_basler_min_exposure(pylon, index=index)

        caps1 = Gst.ElementFactory.make("capsfilter", "capfilter-bin1-%02d" % index)
        caps1.set_property(
            "caps",
            Gst.Caps.from_string(
                f"video/x-raw,width={self.config.width},height={self.config.height},framerate={self.config.framerate}/1,format=YUY2"
            ),
        )
        Gst.Bin.add(nbin, caps1)
        pylon.link(caps1)

        converter = Gst.ElementFactory.make(
            "nvvideoconvert", "nvvideoconvert-bin-%02d" % index
        )
        Gst.Bin.add(nbin, converter)
        caps1.link(converter)

        caps2 = Gst.ElementFactory.make("capsfilter", "capfilter-bin2-%02d" % index)
        caps2.set_property(
            "caps", Gst.Caps.from_string("video/x-raw(memory:NVMM),format=NV12")
        )
        Gst.Bin.add(nbin, caps2)
        converter.link(caps2)

        srcpad = caps2.get_static_pad("src")
        if not srcpad:
            self.logger.error("Unable to create sink pad")

        bin_pad = nbin.add_pad(Gst.GhostPad.new("src", srcpad))
        if not bin_pad:
            self.logger.error("sink bin Failed to add ghost pad in source bin ")
            return None
        self._add_element(nbin)

        self.fps_streams[f"stream{self.num_sources}"] = FPSMonitor(self.num_sources)
        self.num_sources += 1
        return nbin

    def exposure(self, exposure=500):
        for cam in self.cameras.values():
            log.info("camera set exposure %s", exposure)
            cam.set_property("ExposureTime", exposure)
            self.config.exposure = exposure

    def brightness(self, brightness=0.1):
        for cam in self.cameras.values():
            log.info("camera set brightness %s", brightness)
            cam.set_property("AutoTargetBrightness", brightness)
            # self.config.exposure = exposure

    def framesave(self, source_id=0, frame_num=1):
        if self.collector:
            self.collector.framesave(source_id=source_id, frame_num=frame_num)

    def create_fakesource_bin(self, index=0):
        """Create a source GstBin to abstract this bin's content from the rest of the"""

        bin_name = "source-bin-%02d" % index
        nbin = Gst.Bin.new(bin_name)

        bin = Gst.ElementFactory.make("videotestsrc")
        bin.set_property("is_live", True)
        bin.set_property("pattern", 0)
        Gst.Bin.add(nbin, bin)

        converter = Gst.ElementFactory.make("nvvideoconvert")
        Gst.Bin.add(nbin, converter)

        bin.link(converter)

        converter_srcpad = converter.get_static_pad("src")
        if not converter_srcpad:
            self.logger.error("Unable to create queue sink pad")
        #
        bin_pad = nbin.add_pad(Gst.GhostPad.new("src", converter_srcpad))
        if not bin_pad:
            self.logger.error("sink bin Failed to add ghost pad in source bin ")
            return None
        self._add_element(nbin)
        self.num_sources += 1
        return nbin

    def create_source_bin(self, index=0, video_uri=""):
        def _cb_newpad(_, decoder_src_pad, data):
            self.logger.debug("Decodebin pad added")
            caps = decoder_src_pad.get_current_caps()
            gst_struct = caps.get_structure(0)
            gst_name = gst_struct.get_name()
            source_bin = data
            features = caps.get_features(0)

            # Need to check if the pad created by the decodebin is for video and not audio.
            self.logger.debug(f"gstname={gst_name}")
            if gst_name.find("video") != -1:
                # Link the decodebin pad only if decodebin has picked nvidia decoder plugin nvdec_*.
                # We do this by checking if the pad caps contain NVMM memory features.
                self.logger.debug(f"features={features}")
                if features.contains("memory:NVMM"):
                    # Get the source bin ghost pad
                    bin_ghost_pad = source_bin.get_static_pad("src")
                    if not bin_ghost_pad.set_target(decoder_src_pad):
                        self.logger.error(
                            "Failed to link decoder src pad to source bin ghost pad"
                        )
                else:
                    self.logger.error("Decodebin did not pick nvidia decoder plugin.")

        def _decodebin_child_added(child_proxy, obj, name, user_data):
            """
            If the child added to the decodebin is another decodebin, connect to its child-added signal. If the
            child added is a source, set its drop-on-latency property to True.

            :param child_proxy: The child element that was added to the decodebin
            :param Object: The object that emitted the signal
            :param name: The name of the element that was added
            :param user_data: This is a pointer to the data that you want to pass to the callback function
            """
            self.logger.debug(f"Decodebin child added: {name}")
            if name.find("decodebin") != -1:
                obj.connect("child-added", _decodebin_child_added, user_data)

            if "source" in name:
                source_element = child_proxy.get_by_name("source")
                if source_element.find_property("drop-on-latency") is not None:
                    obj.set_property("drop-on-latency", True)

        self.logger.debug("creating source bin index=%s", index)

        # Create a source GstBin to abstract this bin's content from the rest of the pipeline
        bin_name = "source-bin-%02d" % index
        nbin = Gst.Bin.new(bin_name)
        if not nbin:
            self.logger.error("Unable to create source bin")

        uri_decode_bin = self.create_element(
            "uridecodebin", "uri-decode-bin", add=False
        )
        uri_decode_bin.set_property("uri", video_uri)
        # uri_decode_bin.set_property("async-handling", True)
        # uri_decode_bin.set_property("use-buffering", True)

        # Connect to the "pad-added" signal of the decodebin which generates a
        # callback once a new pad for raw data has been created by the decodebin
        uri_decode_bin.connect("pad-added", _cb_newpad, nbin)
        uri_decode_bin.connect("child-added", _decodebin_child_added, nbin)

        # We need to create a ghost pad for the source bin which will act as a proxy
        # for the video decoder src pad. The ghost pad will not have a target right
        # now. Once the decode bin creates the video decoder and generates the
        # cb_newpad callback, we will set the ghost pad target to the video decoder
        # src pad.
        Gst.Bin.add(nbin, uri_decode_bin)
        bin_pad = nbin.add_pad(Gst.GhostPad.new_no_target("src", Gst.PadDirection.SRC))
        if not bin_pad:
            self.logger.error("Failed to add ghost pad in source bin")
            return None

        self._add_element(nbin)

        self.fps_streams[f"stream{index}"] = FPSMonitor(index)
        self.num_sources += 1
        self.config.num_sources = self.num_sources
        return nbin

    # def _create_filesrc(self, filename):
    #     source = self.create_element("filesrc")
    #     source.set_property("location", filename)

    #     h264parser = self.create_element("h264parse")
    #     source.link(h264parser)
    #     decoder = self.create_element("nvv4l2decoder")
    #     h264parser.link(decoder)
    #     return decoder

    def create_convert(self, format="RGBA"):
        """
        from NV12 to 'format' (easier to work with in Python if RGBA)
        """
        self.create_element("nvvideoconvert")
        self.create_caps(f"video/x-raw(memory:NVMM), format={format}")

    def create_screen_out(self, sync=False):
        videorenderer = self.create_element(
            "nv3dsink" if is_aarch64() else "nveglglessink",
        )
        if self.config.new_nvstreammux:
            videorenderer.set_property("async", 1 if self.config.key else 0)
            videorenderer.set_property("sync", sync)
        self.elements_out.append(videorenderer)
        return videorenderer

    def create_fake(self, link=True):
        if link:
            fake = self.create_element("fakesink")
        else:
            fake = self._element("fakesink")
        fake.set_property("enable-last-sample", 0)
        fake.set_property("sync", 0)
        if self.config.new_nvstreammux:
            fake.set_property("async", 0)
        self.elements_out.append(fake)
        return fake

    def create_jpeg_out(self, source_id=0):
        outdir = self.config.save_frames or "/tmp/frames"
        if not os.path.exists(outdir):
            os.makedirs(outdir)

        queue = self._element("queue")  # necessario su jetson per registrare i 2 flussi
        nvconvert = self._element(
            "nvvideoconvert", f"nvvideoconvert_jpeg_{source_id}", nolink=True
        )
        queue.link(nvconvert)
        if is_aarch64():
            nvconvert.set_property("nvbuf-memory-type", 4)
        jpeg = self._element("nvjpegenc", f"nvjpegenc_jpeg_{source_id}", nolink=True)
        nvconvert.link(jpeg)
        jpeg.set_property("quality", self.config.jpeg_quality)
        multi = self._element(
            "multifilesink", f"multifilesink_jpeg_{source_id}", nolink=True
        )
        jpeg.link(multi)
        multi.set_property("location", f"{outdir}/{source_id}_%07d.jpg")
        multi.set_property("max-files", self.config.max_frames)
        multi.set_property("post-messages", True)
        if self.config.new_nvstreammux:
            multi.set_property("sync", False)
            multi.set_property("async", 0)
        # multi.set_property("index", 1)
        # multi.set_property("sync", True) # triplica il tempo della pipeline
        self.elements_out.append(multi)
        return queue

    def create_streammux(self, width=0, height=0, live=False):
        streammux = self.create_element("nvstreammux", "stream-muxer")
        if width:
            self.config.width = width
        if height:
            self.config.height = height
        self.streammux = streammux
        streammux.set_property("batch-size", self.num_sources)

        if not self.config.new_nvstreammux:
            # old nvstreammux
            streammux.set_property("width", self.config.width)
            streammux.set_property("height", self.config.height)
            streammux.set_property(
                "batched-push-timeout", 1000 * 1000 / self.config.framerate
            )
            streammux.set_property("live-source", live)

        # streammux.set_property("sync-inputs", True)
        # streammux.set_property("max-latency", 500000) # 1ms
        return streammux

    def create_streamdemux(self):
        streamdemux = self.create_element(
            "nvstreamdemux", "stream-demuxer", "Stream demux"
        )
        self.streamdemux = streamdemux
        return streamdemux

    def setFramerate(self, framerate: float):
        self.config.framerate = framerate
        self.streammux.set_property(
            "batched-push-timeout", 1000 * 1000 / self.config.framerate
        )

    def create_tracker(self, tracker_config_path):
        tracker = self.create_element("nvtracker")

        tracker.set_property("tracker-width", self.config.tracker_width)
        tracker.set_property("tracker-height", self.config.tracker_height)
        tracker.set_property("gpu-id", 0)
        tracker.set_property(
            "ll-lib-file",
            "/opt/nvidia/deepstream/deepstream/lib/libnvds_nvmultiobjecttracker.so",
        )
        tracker.set_property(
            "ll-config-file",
            asset.load("mev_yolodeep:assets/" + self.config.tracker_config).filename,
        )

        return tracker

    def create_tiler(self):
        tiler = self.create_element("nvmultistreamtiler", "nvtiler", "Tiler")
        tiler_rows = int(math.sqrt(self.num_sources))
        tiler_columns = int(math.ceil((1.0 * self.num_sources) / tiler_rows))
        tiler.set_property("rows", tiler_rows)
        tiler.set_property("columns", tiler_columns)
        if self.config.save:  # manteniamo la dimensione massina
            tiler.set_property("width", 2 * self.config.width * self.config.zoom)
            tiler.set_property("height", self.config.zoom * self.config.height)
        else:
            tiler.set_property("width", self.config.width * self.config.zoom)
            tiler.set_property(
                "height", self.config.zoom * self.config.height / self.num_sources
            )
        return tiler

    def create_caps(self, caps="video/x-raw(memory:NVMM), format=RGBA"):
        capsfilter = self.create_element("capsfilter")
        capsfilter.set_property("caps", Gst.Caps.from_string(caps))
        return capsfilter

    def caps(self, caps="video/x-raw(memory:NVMM), format=RGBA"):
        capsfilter = self._element("capsfilter")
        capsfilter.set_property("caps", Gst.Caps.from_string(caps))
        return capsfilter

    # def _create_videorate(self, rate=0.1, drop=False):
    #     videorate = self.create_element("videorate")
    #     videorate.set_property("rate", rate)
    #     videorate.set_property("drop-only", False)
    #     return videorate

    def create_file_out(self, source_id=-1):
        if source_id >= 0:
            sink = self._element("nvvideoencfilesinkbin")
            base, ext = os.path.splitext(self.config.save)
            filename = f"{base}_{source_id}{ext}"
        else:
            sink = self.create_element("nvvideoencfilesinkbin")
            filename = self.config.save
        sink.set_property("bitrate", self.config.bitrate)
        sink.set_property("codec", self.config.codec)
        sink.set_property("container", "mkv")
        sink.set_property("enc-type", 0)  # 0:hw, 1:sw
        sink.set_property("profile", "main")
        sink.set_property("output-file", filename)
        if self.config.new_nvstreammux:
            sink.set_property("sync", True)
        log.info("saving to %s", filename)
        return sink

    def create_rtsp_out(self):
        sink = self.create_element("nvrtspoutsinkbin")
        sink.set_property("bitrate", self.config.bitrate)
        # sink.set_property("iframeinterval", 3)
        # sink.set_property("qos", 0)
        # sink.set_property("sync", False)
        sink.set_property("codec", self.config.codec)
        sink.set_property("rtsp-port", self.config.rtsp_port)
        sink.set_property("profile", "main")
        if self.config.new_nvstreammux:
            sink.set_property("sync", False)
            sink.set_property("async-handling", 0)
        return sink

    def setup(self):
        pass

    @staticmethod
    def _link_sequential(elements: list):
        for i in range(0, len(elements) - 1):
            elements[i].link(elements[i + 1])

    def _link_elements(self, with_queue=False):
        self.logger.info(f"Linking elements in the Pipeline: {self}")

        source_bins = []
        for element in self.elements:
            if "source-bin" in element.name:  # or "pylonsrc" in element.name:
                source_bins.append(element)

        for i in range(len(self.elements) - 1, 0, -1):
            element_sink = self.elements[i]
            if "source-bin-" in element_sink.name:
                continue

            if element_sink.name == "stream-muxer":
                for source_bin in source_bins:
                    index = int(source_bin.name.split("-")[-1])
                    srcpad = source_bin.get_static_pad("src")

                    # versione con coda che (non) risolve il freeze della pipeline con due flussi video
                    if with_queue:
                        queue = self._element("queue")
                        queue_srcpad = queue.get_static_pad("src")
                        queue_sinkpad = queue.get_static_pad("sink")
                        srcpad.link(queue_sinkpad)
                        sinkpad = element_sink.get_request_pad(f"sink_{index}")
                        queue_srcpad.link(sinkpad)
                        log.debug(
                            "linking source-bin %s -> queue -> streammux %s",
                            source_bin.name,
                            element_sink.name,
                        )
                    else:
                        # versione senza coda
                        sinkpad = element_sink.get_request_pad(f"sink_{index}")
                        srcpad.link(sinkpad)

                        log.debug(
                            "linking source-bin %s -> streammux %s",
                            source_bin.name,
                            element_sink.name,
                        )
            else:
                element_source = self.elements[i - 1]
                log.debug("linking %s -> %s", element_source.name, element_sink.name)
                element_source.link(element_sink)

    # def _write_osd_analytics(
    #     self, batch_meta, l_frame_meta: List, ll_obj_meta: List[List], info=None
    # ):
    #     obj_counter = defaultdict(int)

    #     for frame_meta, l_obj_meta in zip(l_frame_meta, ll_obj_meta):
    #         frame_number = frame_meta.frame_num
    #         num_rects = frame_meta.num_obj_meta

    #         for obj_meta in l_obj_meta:
    #             obj_counter[obj_meta.class_id] += 1
    #             obj_meta.rect_params.border_color.set(0.0, 0.0, 1.0, 0.5)

    #         display_meta = pyds.nvds_acquire_display_meta_from_pool(batch_meta)
    #         display_meta.num_labels = 1
    #         py_nvosd_text_params = display_meta.text_params[0]
    #         py_nvosd_text_params.display_text = "Frame Number={} Number of Objects={} Vehicle_count={} Person_count={}".format(
    #             frame_number,
    #             num_rects,
    #             obj_counter[2],
    #             obj_counter[0],
    #         )

    #         py_nvosd_text_params.x_offset = 10
    #         py_nvosd_text_params.y_offset = 12
    #         py_nvosd_text_params.font_params.font_name = "Serif"
    #         py_nvosd_text_params.font_params.font_size = 10
    #         py_nvosd_text_params.font_params.font_color.set(1.0, 1.0, 1.0, 1.0)
    #         py_nvosd_text_params.set_bg_clr = 1
    #         py_nvosd_text_params.text_bg_clr.set(0.0, 0.0, 0.0, 1.0)

    #         self.logger.info(pyds.get_string(py_nvosd_text_params.display_text))
    #         pyds.nvds_add_display_meta_to_frame(frame_meta, display_meta)

    #         self.fps_streams[f"stream{frame_meta.pad_index}"].get_fps()

    def _calculate_crop_score(self, track_id, crop):
        score = crop.size
        num_detections = len(self.track_scores[track_id])

        # Penalize entry frames
        if num_detections <= 15:
            score -= 1000

        return score

    def _save_crops(
        self, frames, _, l_frame_meta: List, ll_obj_meta: List[List], info=None
    ):
        self.logger.info(f"Saving crops to '{os.path.realpath(CROPS_DIR)}'")
        for frame, frame_meta, l_obj_meta in zip(frames, l_frame_meta, ll_obj_meta):
            frame_copy = np.array(frame, copy=True, order="C")
            frame_copy = cv2.cvtColor(frame_copy, cv2.COLOR_RGBA2BGRA)

            for obj_meta in l_obj_meta:
                track_id = obj_meta.object_id
                x1, y1, x2, y2 = rect_params_to_coords(obj_meta.rect_params)
                crop = frame_copy[y1:y2, x1:x2]
                crop_score = self._calculate_crop_score(track_id, crop)

                if not self.track_scores[track_id] or crop_score > max(
                    self.track_scores[track_id]
                ):
                    crop_dir = os.path.join(
                        CROPS_DIR,
                        f"src_{frame_meta.source_id}",
                        f"obj_{obj_meta.object_id}_cls_{obj_meta.class_id}",
                    )
                    os.makedirs(crop_dir, exist_ok=True)
                    for f in os.listdir(crop_dir):
                        os.remove(os.path.join(crop_dir, f))
                    crop_path = os.path.join(
                        crop_dir, f"frame_{frame_meta.frame_num}.jpg"
                    )
                    cv2.imwrite(crop_path, crop)
                    self.logger.debug(f"Saved crop to '{crop_path}'")

                self.track_scores[track_id].append(crop_score)

    # def _nvjpegencode_init(self, quality=90):
    #     print("NVJPEGENCODE INITIALIZED")
    #     self.create_element("nvvideoconvert")
    #     jpeg = self.create_element("nvjpegenc")
    #     jpeg.set_property("quality", quality)
    #     jpeg_sink = jpeg.get_static_pad("sink")
    #     return jpeg, jpeg_sink

    # def _nvjpeg_save(self, jpeg, frame_data, sink, outdir):
    #     print("NVJPEGENCODE TRIED TO SAVE")
    #     if not os.path.exists(outdir):
    #         os.makedirs(outdir)
    #     temp_buffer = Gst.Buffer.new()
    #     Gst.Buffer.copy_into(temp_buffer, frame_data.get_buffer(), None, True)
    #     # pyds.gst_buffer_copy_region(
    #     #    temp_buffer, frame_data.get_buffer(), 0, 0, 1
    #     # )  # Copy entire frame
    #     pyds.gst_pad_push(jpeg_sinkpad, temp_buffer)

    def _probe_fn_wrapper(self, _, info, probe_fn, get_frames=False):
        # nvjpegenc = False
        # if nvjpegenc:
        #     jpeg, jpeg_sink = self._nvjpegencode_init()

        gst_buffer = info.get_buffer()
        if not gst_buffer:
            self.logger.error("Unable to get GstBuffer")
            return

        frames = []
        l_frame_meta = []
        ll_obj_meta = []
        batch_meta = pyds.gst_buffer_get_nvds_batch_meta(hash(gst_buffer))
        l_frame = batch_meta.frame_meta_list
        while l_frame is not None:
            try:
                frame_meta = pyds.NvDsFrameMeta.cast(l_frame.data)
            except StopIteration:
                break

                # frame = pyds.get_nvds_buf_surface(hash(gst_buffer), frame_meta.batch_id)
                # frames.append(frame)

            # if self.config.experimental:
            #     n_frame = pyds.get_nvds_buf_surface(
            #         hash(gst_buffer), frame_meta.batch_id
            #     )
            #     self._cache_frame(n_frame)

            l_frame_meta.append(frame_meta)
            l_obj_meta = []

            l_obj = frame_meta.obj_meta_list
            while l_obj is not None:
                try:
                    obj_meta = pyds.NvDsObjectMeta.cast(l_obj.data)
                except StopIteration:
                    break

                l_obj_meta.append(obj_meta)

                # if nvjpegenc:
                #     if frame_meta.pad_index % 30 == 0:
                #         # if (frame_meta.pad_index % 30 == 0) and (0.05 < obj_meta.confidence < 0.5):
                #         print("----- TRYING TO SAVE FRAME -----")
                #         # Getting Image data using nvbufsurface
                #         # the input should be address of buffer and batch_id
                #         # n_frame = pyds.get_nvds_buf_surface(
                #         #    hash(gst_buffer), frame_meta.batch_id
                #         # )                        # )                        # )                        # )                        # )                        # )                        # )                        # )                        # )                        # )                        # )

                #         try:
                #             self._nvjpeg_save(
                #                 jpeg, frame_meta, jpeg_sink, "/tmp/context_img"
                #             )
                #             if is_aarch64():  # If Jetson, since the buffer is mapped to CPU for retrieval, it must also be unmapped
                #                 pyds.unmap_nvds_buf_surface(
                #                     hash(gst_buffer), frame_meta.batch_id
                #                 )  # The unmap call should be made after operations with the original array are complete.
                #                 #  The original array cannot be accessed after this call.                            #  The original array cannot be accessed after this call.                            #  The original array cannot be accessed after this call.                            #  The original array cannot be accessed after this call.                            #  The original array cannot be accessed after this call.                            #  The original array cannot be accessed after this call.                            #  The original array cannot be accessed after this call.                            #  The original array cannot be accessed after this call.                            #  The original array cannot be accessed after this call.                            #  The original array cannot be accessed after this call.                            #  The original array cannot be accessed after this call.
                #         except Exception as e:
                #             print(f"COULD NOT SAVE BECAUSE {e}")

                try:
                    l_obj = l_obj.next
                except StopIteration:
                    break

            ll_obj_meta.append(l_obj_meta)

            try:
                l_frame = l_frame.next
            except StopIteration:
                break

            ## --- ⬇ is this necessary if get_frames ? ⬇ --- ##
            if is_aarch64():  # If Jetson, since the buffer is mapped to CPU for retrieval, it must also be unmapped
                pyds.unmap_nvds_buf_surface(hash(gst_buffer), frame_meta.batch_id)
            ## --- ⬆ is this necessary if get_frames ? ⬆ --- ##

        if get_frames:
            probe_fn(frames, batch_meta, l_frame_meta, ll_obj_meta, info=info)
        else:
            probe_fn(batch_meta, l_frame_meta, ll_obj_meta, info=info)

        return Gst.PadProbeReturn.OK

    def _wrap_probe(self, probe_fn):
        get_frames = "frames" in signature(probe_fn).parameters
        return partial(self._probe_fn_wrapper, probe_fn=probe_fn, get_frames=get_frames)

    @staticmethod
    def _get_static_pad(element, pad_name: str = "sink"):
        pad = element.get_static_pad(pad_name)
        if not pad:
            raise AttributeError(f"Unable to get {pad_name} pad of {element.name}")

        return pad

    def _add_probes(self):
        if self.tiler:
            tiler_sinkpad = self._get_static_pad(self.tiler)

            if self.enable_osd and self.write_osd_analytics:
                tiler_sinkpad.add_probe(
                    Gst.PadProbeType.BUFFER, self._wrap_probe(self._write_osd_analytics)
                )

        if self.save_crops:
            tiler_sinkpad.add_probe(
                Gst.PadProbeType.BUFFER, self._wrap_probe(self._save_crops)
            )

    def add_inference_probe(self, sinkpad):
        sinkpad.add_probe(Gst.PadProbeType.BUFFER, self._wrap_probe(self._on_inference))
        self.collector = MainCollector()

    def _display_info(self, batch_meta, frame_meta):
        display_meta = pyds.nvds_acquire_display_meta_from_pool(batch_meta)
        display_meta.num_labels = 1
        py_nvosd_text_params = display_meta.text_params[0]
        py_nvosd_text_params.display_text = f"{frame_meta.frame_num}"

        py_nvosd_text_params.x_offset = int(frame_meta.source_frame_width / 2 + 100)
        py_nvosd_text_params.y_offset = 0
        py_nvosd_text_params.font_params.font_name = "Serif"
        py_nvosd_text_params.font_params.font_size = 20
        py_nvosd_text_params.font_params.font_color.set(0, 0, 0, 1.0)
        py_nvosd_text_params.set_bg_clr = 1
        py_nvosd_text_params.text_bg_clr.set(1.0, 1.0, 1.0, 1.0)

        # self.logger.info(pyds.get_string(py_nvosd_text_params.display_text))
        pyds.nvds_add_display_meta_to_frame(frame_meta, display_meta)

    def _on_inference(
        self, batch_meta, l_frame_meta: List, ll_obj_meta: List[List], info=None
    ):
        # gst_buffer = info.get_buffer()
        for frame_meta, l_obj_meta in zip(l_frame_meta, ll_obj_meta):
            # if self.config.show or self.config.save or self.config.rtsp:
            # get_nvds_buf_Surface only support RGBA color format, in pipeline add video/x-raw(memory:NVMM), format=RGBA converter
            #    frame = pyds.get_nvds_buf_surface(hash(gst_buffer), frame_meta.batch_id)
            # showObject(frame_meta)
            log.debug("frame %d-%d", frame_meta.source_id, frame_meta.frame_num)

            self._display_info(batch_meta, frame_meta)
            # self.collector.addFrame(frame_meta.frame_num, frame_meta.source_id)
            # log.debug(self.collector)

            # if (
            #     self.config.gps_sync
            #     and frame_meta.source_id == 0
            #     and frame_meta.frame_num % self.config.framerate == 0
            # ):
            #     self._emitForGps(frame_meta.source_id, frame_meta.frame_num)

            detections = []
            for obj_meta in l_obj_meta:
                # showObject(obj_meta)
                detection = getDetection(
                    obj_meta,
                    frame_meta,
                    info=info,
                )
                detections.append(detection)

                # if self.config.show or self.config.save or self.config.rtsp:
                #     frame = detection.show(frame)

            try:
                self.fps_streams[f"stream{frame_meta.pad_index}"].get_fps()
            except Exception:
                pass
            # if self.config.save_context and self.collector:
            #     if frame_meta.frame_num > 0 and frame_meta.frame_num % 60 == 0:
            #         self.collector.framesave(
            #             frame_meta.source_id, frame_meta.frame_num - 10
            #         )

            self.on_detections(
                detections,
                source_id=frame_meta.source_id,
                frame_num=frame_meta.frame_num,
            )

    # def _emitForGps(self, source_id, counter):
    #     """
    #     serve in fase di sviluppo per sincronizzare il GPS generato anche quando si rallentano i frames
    #     """
    #     try:
    #         self.sock_emit.sendto(
    #             struct.pack("il", source_id, counter),
    #             (self.config.socket_ip_gps, self.config.socket_port_gps),
    #         )
    #     except Exception:
    #         log.warn(
    #             "failed emitForGps to %s:%s",
    #             self.config.socket_ip_gps,
    #             self.config.socket_port_gps,
    #         )

    ## --- ⬇  frame buffer cache cv2 CUDA ⬇ --- ##
    def _cache_frame(self, n_frame):
        GPU_raw = cv2.cuda_GpuMat()
        GPU_raw.upload(n_frame)
        warp_matrix = np.array(
            [
                [-1.73323209e00, 3.06689479e00, 2.62390280e03],
                [-6.86363713e-16, 5.28105496e00, -9.71431119e02],
                [-5.07596244e-19, 3.19468212e-03, 1.00000000e00],
            ]
        )
        im_res = cv2.cuda.warpPerspective(GPU_raw, warp_matrix, (1920, 1080))
        GPU_frame = cv2.cuda.resize(im_res, (717, 403), interpolation=cv2.INTER_LINEAR)
        GPU_crop = cv2.cuda_GpuMat(
            GPU_frame, (78, 34, 560, 336)
        )  # (top corner X, top corner Y, width X, width Y)
        cuMat = cv2.cuda_GpuMat(336, 560, cv2.CV_32FC4)
        GPU_crop_scale = GPU_crop.convertTo(
            cv2.CV_32FC4, 1 / 255, cv2.cuda.Stream_Null(), cuMat
        )
        frame = np.zeros([336, 560, 4])
        frame = GPU_crop_scale.download(frame)
        print(f"FRAME GOT HERE? {frame}")
        # cv2.imshow(frame)

    ## --- ⬆  buffer cache cv2 CUDA ⬆  --- ##

    def on_detections(
        self, detections: List[Detection], source_id: int = 0, frame_num: int = -1
    ):
        self.collector.add(detections, source_id=source_id, frame_num=frame_num)

    def release(self):
        """Release resources and cleanup."""
        pass

    def print_stats(self):
        for stream in self.fps_streams.values():
            log.debug(stream.print_data())
        return True

    def _readKeyboard(self):
        key = cv2.waitKey(0) & 0xFF
        if key == 27 or key == ord("q"):
            self.exit()

    def step(self):
        """
        Event.step (Format format, uint64 amount, double rate, bool flush, bool intermediate)
        """
        if self.config.key:
            self._readKeyboard()
        for element in self.elements_out:
            element.send_event(
                Gst.Event.new_step(Gst.Format.BUFFERS, 1, 1, True, False)
            )
        return True

    # async def setWamp(self, wampclient):
    #     self.wampclient = wampclient

    def debug(self):
        for i in self.cameras:
            cam = self.cameras[i]
            exposure = cam.get_property("ExposureTime")
            gain = cam.get_property("Gain")
            log.info("camera %d: exposure=%s gain=%s", i, exposure, gain)
        return True

    def starting(self):
        self.logger.info("starting pipeline")
        wamp.publish(
            "yolodeep.start",
            plates_truth=self.config.plates_truth,
            gpx=self.config.gpx,
        )

    def run(self):
        if self.config.wamp:
            import gbulb

            gbulb.install()
            self.loop = asyncio.get_event_loop()
        else:
            self.loop = GLib.MainLoop()

        Gst.debug_bin_to_dot_file(self.pipeline, Gst.DebugGraphDetails.ALL, "pipeline")
        bus = self.pipeline.get_bus()
        bus.add_signal_watch()

        bus.connect("message", self.bus_call, self.loop)

        if self.config.key:
            cv2.namedWindow("main", cv2.WINDOW_AUTOSIZE)

        # GLib.timeout_add(1000, self.debug)

        if self.config.step:
            self.pipeline.set_state(Gst.State.PAUSED)
            GLib.timeout_add(self.config.step, self.step)
        elif self.config.pause:
            self.pipeline.set_state(Gst.State.PAUSED)
        else:
            self.pipeline.set_state(Gst.State.PLAYING)
        GLib.timeout_add(1000, self.starting)

        start = time.time()
        try:
            if not self.config.dummy:
                if self.config.wamp:
                    init_wamp(loop=self.loop, pipeline=self, config=self.config)
                    self.loop.run_forever()
                else:
                    self.loop.run()
        except Exception as e:
            self.logger.warning("stopping pipeline: %s", e)
        finally:
            log.info("elapsed: %s", time.time() - start)
            for stream in self.fps_streams.values():
                stream.get_fps(show=True)
            # time.sleep(999)
            wamp.save()
            self.exit()

    def exit(self):
        self.logger.debug("exiting pipeline")
        self.pipeline.set_state(Gst.State.NULL)
        self.release()
        try:
            asyncio.get_running_loop().stop()
        except Exception:
            pass
        try:
            self.loop.stop()
        except Exception:
            pass
        try:
            self.loop.quit()  # da rimuovere quando wamp sarà attivo
        except Exception:
            pass
