import asyncio
import logging
import sys
import time
import yaml

from autobahn.asyncio.wamp import ApplicationRunner, ApplicationSession

log = logging.getLogger(__name__)


def init_wamp(loop, pipeline, config):
    WampClient.config = config
    runner = ApplicationRunner(
        url=config.ws,
        realm=config.realm,
        # serializers=[JsonSerializer()],
    )
    log.info(
        "connecting to wamp router %s on realm %s",
        config.ws,
        config.realm,
    )

    def check():
        if not WampClient._is_connected:
            log.warning("no connection made, exit")
            asyncio.get_event_loop().stop()

    asyncio.get_event_loop().call_later(5, check)

    loop.create_task(runner.run(WampClient.init(pipeline), start_loop=False))


class WampClient(ApplicationSession):
    pipeline = None
    _is_connected: bool = False
    config = None

    async def onJoin(self, details):
        log.info("wamp session attached to realm")
        WampClient._is_connected = True
        await self.register(self.ping, "ping")
        await self.subscribe(self.pipeline.step, "pipeline.step")
        await self.subscribe(self.pipeline.exposure, "camera.exposure")
        await self.subscribe(self.pipeline.brightness, "camera.brightness")
        await self.subscribe(self.pipeline.framesave, "frame.save")

        Wamp.session = self

    async def ping(self):
        return {"message": "pong", "time": time.time()}

    def close(self, force=False):
        try:
            asyncio.get_running_loop().stop()
        except Exception:
            log.warning("no asyncio loop detected")
        if force:
            log.warning("exit force requested... bye...")
            sys.exit(0)

    @staticmethod
    def init(pipeline):
        WampClient.pipeline = pipeline
        return WampClient

    def onLeave(self, details):
        log.warning("WAMP session ended")
        # self.count += 1
        # if self.count < 3:
        #    self.join(u"realm1")
        # else:
        self.disconnect()

    def onDisconnect(self):
        log.warning("Wamp transport disconnected.")
        # try:
        #     for task in asyncio.Task.all_tasks():
        #         log.warning("Cancelling waiting task...")
        #         task.cancel()
        # except Exception:
        #     for task in asyncio.all_tasks():
        #         log.warning("Cancelling waiting task...")
        #         task.cancel()
        asyncio.get_event_loop().stop()


class Wamp:
    session = WampClient()

    def __init__(self):
        self._messages = []
        self._last_message_time = 0

    def publish(self, uri, **kwargs):
        if self.session._is_connected:
            log.debug("wamp publish: %s %s", uri, kwargs)
            self.session.publish(uri, **kwargs)
            if self.session.config.save_wamp:
                self._save(uri, kwargs)

    def _save(self, uri, payload):
        now = time.time()
        if not self._last_message_time:
            self._last_message_time = now
        self._messages.append(
            {"uri": uri, "payload": payload, "delay": now - self._last_message_time}
        )
        self._last_message_time = now

    def save(self):
        if self.session._is_connected and self.session.config.save_wamp:
            with open(self.session.config.save_wamp, "w") as outfile:
                yaml.dump(self._messages, outfile, default_flow_style=False)
                log.info("saved wamp messages to %s", self.session.config.save_wamp)


wamp = Wamp()
